-- Copyright 2016, 2017, 2018 Cumulus Networks, Inc.  All Rights reserved.
--
-- This software is subject to the Cumulus Networks End User License Agreement
-- available at the following locations:
--
-- Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
-- Cumulus Linux systems: /usr/share/cumulus/EULA.txt

function zcount_script()
    local key_idx, tmp1, tmp2
    local result = {}
--  redis.log(redis.LOG_WARNING, "zcount begin")
    for key_idx = 1, #KEYS do
        tmp1 = redis.call("ZRANGE", KEYS[key_idx], -1, -1, "WITHSCORES")
        tmp2 = redis.call("ZREVRANGE", KEYS[key_idx], -1, -1, "WITHSCORES")
        result[#result + 1] = {KEYS[key_idx],
                               redis.call("ZCARD", KEYS[key_idx]),
                               tmp2[2],
                               tmp1[2]}
    end
--  redis.log(redis.LOG_WARNING, "zcount end")
    return result
end

function count_script()
  local tmp, entry, idx
  local result = 0
  local total_mem = 0
  local match = KEYS[1]
  local cursor = KEYS[2]
  local keys, val_json
  local start_time = ARGV[1]
  local objtype = ARGV[2]
  local separator = ARGV[3]
  local class_name = ARGV[4]
  local do_count
--redis.log(redis.LOG_WARNING, "count begin")
  tmp = redis.call("HSCAN", class_name, cursor, "MATCH", match, "COUNT", 1000)
  cursor = tonumber(tmp[1])
  keys = tmp[2]
  if keys ~= nil then
     for idx = 1, #keys do
        do_count = 0
        if objtype == "Link" then
          -- these are the fields we don't display
          if not string.match(keys[idx], "-v0"..separator) then
            do_count = 1
          end
        elseif objtype == "Address" then
          -- these are the fields we don't display
          if not string.match(keys[idx], separator.."127.0.0.1"..separator) and
            not string.match(keys[idx], separator.."::1"..separator) and
          not string.match(keys[idx], "-v0"..separator.."fe80::") then
            do_count = 1
          end
        elseif objtype == "Route" then
          -- these are the fields we don't display
          if not string.match(keys[idx], separator.."127.") and
            not string.match(keys[idx], separator.."fe80:") and
            not string.match(keys[idx], separator.."::1/128") and
            not string.match(keys[idx], separator.."ff00::/8") and
            not string.match(keys[idx], "Route"..separator.."(.+)"..separator.."3"..separator) and
            not string.match(keys[idx], separator.."(.+)/128"..separator.."1"..separator.."(.+)"..separator.."1"..separator.."([0-1])"..separator.."kernel")
          then
            do_count = 1
          end
        elseif objtype == "Neighbor" then
          if not string.match(keys[idx], separator.."lo"..separator) and
          not string.match(keys[idx], separator.."None"..separator) then
            do_count = 1
          end
        elseif objtype == "MacFdb" then
          if not string.match(keys[idx], separator.."00:00:00:00:00:00"..separator) and
            not string.match(keys[idx], separator.."0$")
          then
            do_count = 1
          end
        else
          do_count = 1
        end

        if do_count == 1 then
          entry = redis.call("ZREVRANGEBYSCORE", keys[idx], start_time, 0, "LIMIT", 0, 1)
          if entry[1] then
            val_json = cjson.decode(entry[1])
            if val_json["active"] == "1" then
              result = result + 1
            end
          end
        end
     end
  end
--redis.log(redis.LOG_WARNING, "count end")
  return {cursor, result}
end

function delregex_script()
    local new_score = ARGV[1]
    local purge = ARGV[2]
    local tmp, score, val, val_json
    local updated = {}
--  redis.log(redis.LOG_WARNING, "delregex begin ", #KEYS, KEYS[1])
    for idx = 1, #KEYS do
        tmp = redis.call("ZRANGE", KEYS[idx], -1, -1, "WITHSCORES")
        val = tmp[1]
        score = tmp[2]
        if val ~= nil then
            if purge == 1 then
                redis.call("DEL", KEYS[idx])
            else
                val_json = cjson.decode(val)
                if val_json["active"] ~= "0" then
                    val_json["active"] = "0"
                    val_json["timestamp"] = new_score
                    redis.call("ZADD", KEYS[idx], new_score, cjson.encode(val_json))
                    updated[#updated + 1] = {KEYS[idx], score, val}
                    updated[#updated + 1] = {KEYS[idx], new_score, cjson.encode(val_json) }
                else
                    tmp = redis.call("ZRANGE", KEYS[idx], -2, -2, "WITHSCORES")
                    if tmp[1] ~= nil then
                        updated[#updated + 1] = {KEYS[idx], tmp[2], tmp[1]}
                    end
                    updated[#updated + 1] = {KEYS[idx], score, val}
                end
            end
        end
    end
--  redis.log(redis.LOG_WARNING, "delregex end ", #updated)
    return updated
end

function gc_script()
    local now = math.floor(ARGV[1])
    local freed = 0
    local tmp, tmp1, tmp2, key1, idx, idx1
--  redis.log(redis.LOG_WARNING, "gc begin")
    for index = 1, #KEYS do
        local counters = {}
        tmp = redis.call("ZRANGE", KEYS[index], 0, 10000, "WITHSCORES")
        -- always preserve the latest, the last index in the array is the latest
        -- score
        local latest = 0
        if tmp[#tmp] then
           latest=math.ceil(tmp[#tmp])
        else
           latest=now
        end

        for idx=1, #tmp, 2 do
            local score = tmp[idx + 1]
            local ago = now - score
            if ago > 60 * 60 then
               tmp1 = math.floor(ago / (60 * 60) + 1)
               tmp2 = math.floor(now - (tmp1 - 1) * (60 * 60))
               if tmp2 >= latest then
                  tmp2 = latest - 1
               end
               key1 = tmp1..'h'
            elseif ago > 60 then
               tmp1 = math.floor(ago / 60 + 1)
               tmp2 = math.floor(now - (tmp1 - 1) * 60)
               if tmp2 >= latest then
                  tmp2 = latest - 1
               end
               key1 = tmp1..'m'
            else
               tmp1 = math.floor(ago + 1)
               tmp2 = math.floor(now - (tmp1 - 1))
               if tmp2 >= latest then
                  tmp2 = latest - 1
               end
               key1 = tmp1..'s'
            end
            if counters[key1] ~= 1 then
               freed = freed + redis.call("ZCOUNT", KEYS[index], '('..score, tmp2)
               redis.call("ZREMRANGEBYSCORE", KEYS[index], '('..score, tmp2)
               counters[key1] = 1
            end
        end
    end
--  redis.log(redis.LOG_WARNING, "gc end")
    return freed
end

function keys_script()
    local match = KEYS[1]
    local objtype = ARGV[1]
    local tmp, keys
    local cur = KEYS[2]
    local count = KEYS[3]
    local result = {}
--  redis.log(redis.LOG_WARNING, "keys begin", match, objtype, cur, count)
    repeat
        tmp = redis.call("HSCAN", objtype, cur, "MATCH", match, "COUNT", count)
        cur = tonumber(tmp[1])
        keys = tmp[2]
        if keys ~= nil then
            for idx = 1, #keys, 2 do
                result[#result + 1] = keys[idx]
            end
        end
    until cur == 0 or (count and #result >= tonumber(count))
--  redis.log(redis.LOG_WARNING, "keys end ", #result)
    return {cur, result}
end

function lpm_script()
    -- Perform a LPM against the prefix.
    -- a. Start by constructing a regex assuming keys are of a fixed format.
    -- b. In every iteration of the outer loop, we perform the logical
    --    equivalent of right shifting (by 8/16) the prefix string and
    --    decrementing the mask by 8/16 till
    --    i. all the bits in the prefix string have been zeroed out, or
    --   ii. a LPM for the hostnames passed to this method has been found
    -- c. In the inner loop, we iterate over the Routes in the database
    --    extracting the prefix, mask and hostname using the regex from a.
    --    Subsequently, we match the prefix and mask against the prefix and
    --    mask from b. If a match is found, we extract the score
    --    and value for the key and store it as the LPM for that hostname.
    local hostnames = KEYS
    local match = ARGV[1]
    local input_prefix = ARGV[2]
    local step_down_plen = tonumber(ARGV[3])
    local start_time = ARGV[4]
    local is_ipv6 = tonumber(ARGV[5])
    local origin = tonumber(ARGV[6])
    local seperator = ARGV[7]
	local class_name = ARGV[8]

    -- Regex construction
    local ipv4seg = "%d%d?%d?"
    local ipv6seg = "[0-9a-fA-F]?[0-9a-fA-F]?[0-9a-fA-F]?[0-9a-fA-F]?:?:?"
    local remask = "%d+"
    -- We assume the hostname comes before the IP address
    local reipv4addr = string.format("Route"..seperator.."(.+)"..seperator.."(%s)/(%s)"..seperator.."(%s)"..seperator.."(%s)"..seperator,
                                     string.rep(ipv4seg.."%.", 3)..ipv4seg,
                                     remask,
                                     "%d+", "%d+")
    local reipv6addr = string.format("Route"..seperator.."(.+)"..seperator.."(%s)/(%s)"..seperator.."(%s)"..seperator.."(%s)"..seperator,
                                     string.rep(ipv6seg, 8),
                                     remask,
                                     "%d+", "%d+")

    -- Variable initialization
    local src_prefix_inc, src_prefix_len
    if is_ipv6 == 1 then
        src_prefix_len = 128
        src_prefix_inc = 16
    else
        src_prefix_len = 32
        src_prefix_inc = 8
    end

    local tmp, longest_key, longest_val, longest_score, h, t
    local results = {}
    local found = {}
    local key = match
    local prefix_len = src_prefix_len
    local prefix_str = input_prefix
    while prefix_len >= 0 do
        prefix_str = utils.sub_prefix(prefix_str, is_ipv6)
        if prefix_str == nil then
          break
        end
        local new_key = string.gsub(key, seperator..input_prefix, seperator..prefix_str)
        local cur = 0
        repeat
            tmp = redis.call("HSCAN", class_name, cur, "MATCH", new_key, "COUNT", 1000)
            cur = tonumber(tmp[1])
            local keys = tmp[2]
            local hostname, prefix, mask, rtype, val_json, table, found_key
            if #keys ~= 0 then
                for idx = 1, #keys, 2 do
                    if is_ipv6 == 1 then
                        hostname, prefix, mask, rtype, table = string.match(keys[idx], reipv6addr)
                    else
                        hostname, prefix, mask, rtype, table = string.match(keys[idx], reipv4addr)
                    end
                    mask = tonumber(mask) or 0
                    if table ~= 0 then
                      found_key = hostname.."+"..table
                      -- Broadcast routes are of type 3

                      if (found[found_key] == nil or mask > found[found_key][1])
                        and tonumber(rtype) ~= 3 and mask <= prefix_len
                        and ((is_ipv6 == 1
                                and (((mask == 128 and input_prefix == prefix) or mask ~= 128)))
                            or (is_ipv6 ~= 1 and utils.is_match(prefix, input_prefix, mask)))
                      then
                        longest_key = keys[idx]
                        tmp = redis.call("ZREVRANGEBYSCORE", longest_key, start_time, 0, "WITHSCORES", "LIMIT", 0, 1)
                        if tmp[1] then
			   longest_val = tmp[1]
			   val_json = cjson.decode(longest_val)
			   if val_json["active"] == "1" then
			      longest_score = tmp[2]
			      found[found_key] = {mask, longest_key, longest_val, longest_score}
			   end
                        end
                      end
                    end
                end
            end
        until cur == 0
	if origin == 1 then
	   -- Try and end the cycle earlier
	   if prefix_len == src_prefix_len then
	      prefix_len = 0
	   end
	end
        if prefix_len == 128 then
          prefix_len = step_down_plen
        else
          prefix_len = prefix_len - src_prefix_inc
        end
    end

    for k, v in pairs(found) do
        for idx = 1, #hostnames do
            h, t = string.match(k, "(.*)".."(+)"..("%d+"))
            if h == hostnames[idx] then
                longest_key = v[2]
                longest_val = v[3]
                longest_score = v[4]
                results[#results + 1] = {longest_key, longest_val, longest_score }
                break
            end
        end
    end

    return results
end

function scanregex_script()
    local match = KEYS[1]
    local nmatch = KEYS[2]
    local cur = tonumber(KEYS[3])
	local class = KEYS[4]
    local count = KEYS[5]
    local rangeapi = ARGV[1]
    local rangestart = ARGV[2]
    local rangeend = ARGV[3]
    local first = tonumber(ARGV[4])
    local offset = ARGV[5]
    local limit = ARGV[6]
    local rep = {}
    local tmp, keys
    local totkeys = 0
--  redis.log(redis.LOG_WARNING, "scanr begin", match, cur, count)
    repeat
        tmp = redis.call("HSCAN", class, cur, "MATCH", match, "COUNT", count)
        cur = tonumber(tmp[1])
        keys = tmp[2]
        totkeys = totkeys + #keys
        if #keys ~= 0 then
            for idx = 1, #keys do
                local key_type = redis.call("TYPE", keys[idx])
                if type(key_type) == "table" then
                    -- hacky conversion from table to string
                    for _, tmp in pairs(key_type) do
                        key_type = tmp
                    end
                end
                if nmatch == ".*" or not keys[idx]:find(nmatch) then
                    if key_type == "set" or key_type == "string" then
                        rep[#rep + 1] = {keys[idx], redis.call("GET", keys[idx]), 0}
                    elseif key_type == "hash" then
                        tmp = redis.call("HGETALL", keys[idx])
                        for index = 1, #tmp do
                            rep[#rep + 1] = {keys[idx], tmp[index], 0}
                        end
                    elseif key_type == "zset" then
                        if rangeapi == 'ZREVRANGEBYSCORE' or rangeapi == 'ZRANGEBYSCORE' then
                            tmp = redis.call(rangeapi, keys[idx], rangestart, rangeend, "WITHSCORES", "LIMIT", offset, limit)
                        else
                            tmp = redis.call(rangeapi, keys[idx], rangestart, rangeend, "WITHSCORES")
                        end
                        for index=1, #tmp, 2 do
                            rep[#rep + 1] = {keys[idx], tmp[index], tmp[index + 1]}
                            if first == 1 then
                                break
                            end
                        end
                    end
                end
            end
        end
    until cur == 0 or (count and totkeys >= tonumber(count))
--  redis.log(redis.LOG_WARNING, "scanr end", #rep, totkeys)
    return {cur, rep}
end
